#version 300 es

precision highp float;

const mediump vec3 satLuminanceWeighting = vec3(0.2126, 0.7152, 0.0722);
const mediump vec3 hsLuminanceWeighting = vec3(0.3, 0.3, 0.3);

uniform sampler2D u_Texture;
uniform lowp float u_Contrast;
uniform lowp float u_Brightness;
uniform lowp float u_Saturation;
uniform lowp float u_Shadows;
uniform lowp float u_Highlights;
uniform lowp float u_ShadowsTintIntensity;
uniform lowp vec3 u_ShadowsTintColor;
uniform lowp float u_FadeAmount;
uniform lowp float u_Warmth;
uniform lowp float u_Exposure;
uniform float u_Sharpen;

uniform int u_ApplyCurves;
uniform sampler2D u_CurvesImage;

in vec2 v_LeftTexCoord;
in vec2 v_RightTexCoord;
in vec2 v_TopTexCoord;
in vec2 v_BottomTexCoord;
in vec2 v_TexCoord;

layout(location = 0) out vec4 o_FragColor;

highp vec3 rgbToHsl(highp vec3 color) {
    highp vec3 hsl;
    highp float fmin = min(min(color.r, color.g), color.b);
    highp float fmax = max(max(color.r, color.g), color.b);
    highp float delta = fmax - fmin;
    hsl.z = (fmax + fmin) / 2.0;

    if (delta == 0.0) {
        hsl.x = 0.0;
        hsl.y = 0.0;
    } else {
        if (hsl.z < 0.5) {
            hsl.y = delta / (fmax + fmin);
        } else {
            hsl.y = delta / (2.0 - fmax - fmin);
        }

        highp float deltaR = (((fmax - color.r) / 6.0) + (delta / 2.0)) / delta;
        highp float deltaG = (((fmax - color.g) / 6.0) + (delta / 2.0)) / delta;
        highp float deltaB = (((fmax - color.b) / 6.0) + (delta / 2.0)) / delta;

        if (color.r == fmax) {
            hsl.x = deltaB - deltaG;
        } else if (color.g == fmax) {
            hsl.x = (1.0 / 3.0) + deltaR - deltaB;
        } else if (color.b == fmax) {
            hsl.x = (2.0 / 3.0) + deltaG - deltaR;
        }

        if (hsl.x < 0.0) {
            hsl.x += 1.0;
        } else if (hsl.x > 1.0) {
            hsl.x -= 1.0;
        }
    }
    return hsl;
}

highp float hueToRgb(highp float f1, highp float f2, highp float hue) {
    if (hue < 0.0) {
        hue += 1.0;
    } else if (hue > 1.0) {
        hue -= 1.0;
    }

    highp float res;

    if ((6.0 * hue) < 1.0) {
        res = f1 + (f2 - f1) * 6.0 * hue;
    } else if ((2.0 * hue) < 1.0) {
        res = f2;
    } else if ((3.0 * hue) < 2.0) {
        res = f1 + (f2 - f1) * ((2.0 / 3.0) - hue) * 6.0;
    } else {
        res = f1;
    } return res;
}

highp vec3 hslToRgb(highp vec3 hsl) {
    if (hsl.y == 0.0) {
        return vec3(hsl.z);
    } else {
        highp float f2;
        if (hsl.z < 0.5) {
            f2 = hsl.z * (1.0 + hsl.y);
        } else {
            f2 = (hsl.z + hsl.y) - (hsl.y * hsl.z);
        }
        highp float f1 = 2.0 * hsl.z - f2;
        return vec3(hueToRgb(f1, f2, hsl.x + (1.0/3.0)), hueToRgb(f1, f2, hsl.x), hueToRgb(f1, f2, hsl.x - (1.0/3.0)));
    }
}

lowp vec3 applyLuminanceCurve(lowp vec3 pixel) {
    highp float index = floor(clamp(pixel.z / (1.0 / 200.0), 0.0, 199.0));
    pixel.y = mix(0.0, pixel.y, smoothstep(0.0, 0.1, pixel.z) * (1.0 - smoothstep(0.8, 1.0, pixel.z)));
    pixel.z = texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).a;
    return pixel;
}

lowp vec3 applyRGBCurve(lowp vec3 pixel) {
    highp float index = floor(clamp(pixel.r / (1.0 / 200.0), 0.0, 199.0));
    pixel.r = texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).r;
    index = floor(clamp(pixel.g / (1.0 / 200.0), 0.0, 199.0));
    pixel.g = clamp(texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).g, 0.0, 1.0);
    index = floor(clamp(pixel.b / (1.0 / 200.0), 0.0, 199.0));
    pixel.b = clamp(texture(u_CurvesImage, vec2(1.0 / 200.0 * index, 0)).b, 0.0, 1.0);
    return pixel;
}

lowp vec3 tintRaiseShadowsCurve(lowp vec3 color) {
    return vec3(-0.003671) * pow(color, vec3(3.0)) + vec3(0.3842) * pow(color, vec3(2.0)) + vec3(0.3764) * color + vec3(0.2515);
}

lowp vec3 tintShadows(lowp vec3 texel, lowp vec3 tintColor, lowp float tintAmount) {
    return clamp(mix(texel, mix(texel, tintRaiseShadowsCurve(texel), tintColor), tintAmount), 0.0, 1.0);
}

highp vec3 fadeAdjust(highp vec3 color, highp float fadeVal) {
    return (color * (1.0 - fadeVal)) + ((color + (vec3(-0.9772) * pow(vec3(color), vec3(3.0)) + vec3(1.708) * pow(vec3(color), vec3(2.0)) + vec3(-0.1603) * vec3(color) + vec3(0.2878) - color * vec3(0.9))) * fadeVal);
}

highp float getLuma(highp vec3 rgbP) {
    return (0.299 * rgbP.r) + (0.587 * rgbP.g) + (0.114 * rgbP.b);
}

highp vec3 rgbToYuv(highp vec3 inP) {
    highp float luma = getLuma(inP);

    return vec3(luma, (1.0 / 1.772) * (inP.b - luma), (1.0 / 1.402) * (inP.r - luma));
}

lowp vec3 yuvToRgb(highp vec3 inP) {
    return vec3(1.402 * inP.b + inP.r, (inP.r - (0.299 * 1.402 / 0.587) * inP.b - (0.114 * 1.772 / 0.587) * inP.g), 1.772 * inP.g + inP.r);
}

void main(void) {
    const lowp float toolEpsilon = 0.005;
    lowp vec4 source = texture(u_Texture, v_TexCoord);
    lowp vec4 result = source;

    if (u_ApplyCurves == 1) {
        result = vec4(applyRGBCurve(hslToRgb(applyLuminanceCurve(rgbToHsl(result.rgb / result.a)))), 1.0) * result.a;
    }

    mediump float hsLuminance = dot(result.rgb, hsLuminanceWeighting);
    mediump float shadow = clamp((pow(hsLuminance, 1.0 / u_Shadows) + (-0.76) * pow(hsLuminance, 2.0 / u_Shadows)) - hsLuminance, 0.0, 1.0);
    mediump float highlight = clamp((1.0 - (pow(1.0 - hsLuminance, 1.0 / (2.0 - u_Highlights)) + (-0.8) * pow(1.0 - hsLuminance, 2.0 / (2.0 - u_Highlights)))) - hsLuminance, -1.0, 0.0);
    lowp vec3 hsresult = vec3(0.0, 0.0, 0.0) + ((hsLuminance + shadow + highlight) - 0.0) * ((result.rgb - vec3(0.0, 0.0, 0.0)) / (hsLuminance - 0.0));
    mediump float contrastedLuminance = ((hsLuminance - 0.5) * 1.5) + 0.5;
    mediump float whiteInterp = contrastedLuminance * contrastedLuminance * contrastedLuminance;
    mediump float whiteTarget = clamp(u_Highlights, 1.0, 2.0) - 1.0;
    hsresult = mix(hsresult, vec3(1.0), whiteInterp * whiteTarget);
    mediump float invContrastedLuminance = 1.0 - contrastedLuminance;
    mediump float blackInterp = invContrastedLuminance * invContrastedLuminance * invContrastedLuminance;
    mediump float blackTarget = 1.0 - clamp(u_Shadows, 0.0, 1.0);
    hsresult = mix(hsresult, vec3(0.0), blackInterp * blackTarget);
    result = vec4(hsresult.rgb, result.a);

    result = vec4(clamp(((result.rgb - vec3(0.5)) + u_Brightness + vec3(0.5)), 0.0, 1.0), result.a);

    result = vec4(clamp(((result.rgb / result.a - vec3(0.5)) * u_Contrast + vec3(0.5)), 0.0, 1.0), 1.0) * result.a;

    if (abs(u_FadeAmount) > toolEpsilon) {
        result.rgb = fadeAdjust(result.rgb / result.a, u_FadeAmount) * result.a;
    }

    lowp float satLuminance = dot(result.rgb, satLuminanceWeighting);
    lowp vec3 greyScaleColor = vec3(satLuminance);
    result = vec4(clamp(mix(greyScaleColor, result.rgb, u_Saturation), 0.0, 1.0), result.a);

    if (abs(u_ShadowsTintIntensity)>toolEpsilon) {
        result.rgb = tintShadows(result.rgb, u_ShadowsTintColor, u_ShadowsTintIntensity * 2.0);
    }

    if (abs(u_Exposure)>toolEpsilon) {
        mediump float mag = u_Exposure * 1.045;
        mediump float exppower = 1.0 + abs(mag);

        if (mag < 0.0) {
            exppower = 1.0 / exppower;
        }

        result.r = 1.0 - pow((1.0 - result.r), exppower);
        result.g = 1.0 - pow((1.0 - result.g), exppower);
        result.b = 1.0 - pow((1.0 - result.b), exppower);
    }

    if (abs(u_Warmth)>toolEpsilon){
        highp vec3 yuvVec;

        if (u_Warmth > 0.0 ) {
            yuvVec = vec3(0.1765, -0.1255, 0.0902);
        } else {
            yuvVec = -vec3(0.0588, 0.1569, -0.1255);
        }

        highp vec3 yuvColor = rgbToYuv(result.rgb);
        highp float luma = yuvColor.r;
        highp float curveScale = sin(luma * 3.14159);
        yuvColor += 0.375 * u_Warmth * curveScale * yuvVec;
        result.rgb = yuvToRgb(yuvColor);
    }

    vec3 leftTextureColor = texture(u_Texture, v_LeftTexCoord).rgb;
    vec3 rightTextureColor = texture(u_Texture, v_RightTexCoord).rgb;
    vec3 topTextureColor = texture(u_Texture, v_TopTexCoord).rgb;
    vec3 bottomTextureColor = texture(u_Texture, v_BottomTexCoord).rgb;
    result.rgb = result.rgb * (1.0 + 4.0 * u_Sharpen) - (leftTextureColor + rightTextureColor + topTextureColor + bottomTextureColor) * u_Sharpen;

	o_FragColor = result;
}